package sun.nio.ch;

import java.security.AccessController;
import sun.security.action.GetIntegerAction;
import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionException;
import java.nio.channels.ShutdownChannelGroupException;
import java.nio.channels.AsynchronousChannel;
import sun.misc.InnocuousThread;
import java.nio.channels.CompletionHandler;

class Invoker
{
    private static final int maxHandlerInvokeCount;
    private static final ThreadLocal<Invoker.GroupAndInvokeCount> myGroupAndInvokeCount;
    
    private Invoker() {
    }
    
    static void bindToGroup(AsynchronousChannelGroupImpl asynchronousChannelGroupImpl) {
        Invoker.myGroupAndInvokeCount.set(new GroupAndInvokeCount(asynchronousChannelGroupImpl));
    }
    
    static Invoker.GroupAndInvokeCount getGroupAndInvokeCount() {
        return (Invoker.GroupAndInvokeCount)Invoker.myGroupAndInvokeCount.get();
    }
    
    static boolean isBoundToAnyGroup() {
        return Invoker.myGroupAndInvokeCount.get() != null;
    }
    
    static boolean mayInvokeDirect(Invoker.GroupAndInvokeCount groupAndInvokeCount, AsynchronousChannelGroupImpl asynchronousChannelGroupImpl) {
        return groupAndInvokeCount != null && groupAndInvokeCount.group() == asynchronousChannelGroupImpl && groupAndInvokeCount.invokeCount() < Invoker.maxHandlerInvokeCount;
    }
    
    static <V, A> void invokeUnchecked(CompletionHandler<V, ? super A> completionHandler, A a, V v, Throwable t) {
        if (t == null) {
            completionHandler.completed(v, a);
        }
        else {
            completionHandler.failed(t, a);
        }
        Thread.interrupted();
        if (System.getSecurityManager() != null) {
            Thread currentThread = Thread.currentThread();
            if ((currentThread instanceof InnocuousThread)) {
                Invoker.GroupAndInvokeCount value = (Invoker.GroupAndInvokeCount)Invoker.myGroupAndInvokeCount.get();
                ((InnocuousThread)currentThread).eraseThreadLocals();
                if (value != null) {
                    Invoker.myGroupAndInvokeCount.set(value);
                }
            }
        }
    }
    
    static <V, A> void invokeDirect(Invoker.GroupAndInvokeCount groupAndInvokeCount, CompletionHandler<V, ? super A> completionHandler, A a, V v, Throwable t) {
        groupAndInvokeCount.incrementInvokeCount();
        invokeUnchecked(completionHandler, a, v, t);
    }
    
    static <V, A> void invoke(AsynchronousChannel asynchronousChannel, CompletionHandler<V, ? super A> completionHandler, A a, V v, Throwable t) {
        boolean n = false;
        boolean n2 = false;
        Invoker.GroupAndInvokeCount groupAndInvokeCount = (Invoker.GroupAndInvokeCount)Invoker.myGroupAndInvokeCount.get();
        if (groupAndInvokeCount != null) {
            if (groupAndInvokeCount.group() == ((Groupable)asynchronousChannel).group()) {
                n2 = true;
            }
            if (n2 && groupAndInvokeCount.invokeCount() < Invoker.maxHandlerInvokeCount) {
                n = true;
            }
        }
        if (!(n == 0)) {
            invokeDirect(groupAndInvokeCount, completionHandler, a, v, t);
        }
        else {
            try {
                invokeIndirectly(asynchronousChannel, completionHandler, a, v, t);
            }
            catch (RejectedExecutionException ex) {
                if (!(n2 == 0)) {
                    invokeDirect(groupAndInvokeCount, completionHandler, a, v, t);
                }
                else {
                    throw new ShutdownChannelGroupException();
                }
            }
        }
    }
    
    static <V, A> void invokeIndirectly(AsynchronousChannel asynchronousChannel, final CompletionHandler<V, ? super A> completionHandler, final A a, final V v, final Throwable t) {
        try {
            ((Groupable)asynchronousChannel).group().executeOnPooledThread(new Runnable() {
                @Override
                public void run() {
                    Invoker.GroupAndInvokeCount groupAndInvokeCount = (Invoker.GroupAndInvokeCount)Invoker.myGroupAndInvokeCount.get();
                    if (groupAndInvokeCount != null) {
                        groupAndInvokeCount.setInvokeCount(1);
                    }
                    Invoker.invokeUnchecked(completionHandler, a, v, t);
                }
            });
        }
        catch (RejectedExecutionException ex) {
            throw new ShutdownChannelGroupException();
        }
    }
    
    static <V, A> void invokeIndirectly(final CompletionHandler<V, ? super A> completionHandler, final A a, final V v, final Throwable t, Executor executor) {
        try {
            executor.execute(new Runnable() {
                @Override
                public void run() {
                    Invoker.invokeUnchecked(completionHandler, a, v, t);
                }
            });
        }
        catch (RejectedExecutionException ex) {
            throw new ShutdownChannelGroupException();
        }
    }
    
    static void invokeOnThreadInThreadPool(Groupable groupable, Runnable runnable) {
        Invoker.GroupAndInvokeCount groupAndInvokeCount = (Invoker.GroupAndInvokeCount)Invoker.myGroupAndInvokeCount.get();
        AsynchronousChannelGroupImpl group = groupable.group();
        int n;
        if (groupAndInvokeCount == null) {
            n = 0;
        }
        else {
            n = ((groupAndInvokeCount.group == group) ? 1 : 0);
        }
        try {
            if (n != 0) {
                runnable.run();
            }
            else {
                group.executeOnPooledThread(runnable);
            }
        }
        catch (RejectedExecutionException ex) {
            throw new ShutdownChannelGroupException();
        }
    }
    
    static <V, A> void invokeUnchecked(PendingFuture<V, A> pendingFuture) {
        assert !(!pendingFuture.isDone());
        CompletionHandler<V, Object> handler = pendingFuture.handler();
        if (handler != null) {
            invokeUnchecked(handler, pendingFuture.attachment(), pendingFuture.value(), pendingFuture.exception());
        }
        return;
    }
    
    static <V, A> void invoke(PendingFuture<V, A> pendingFuture) {
        assert !(!pendingFuture.isDone());
        CompletionHandler<V, Object> handler = pendingFuture.handler();
        if (handler != null) {
            invoke(pendingFuture.channel(), handler, pendingFuture.attachment(), pendingFuture.value(), pendingFuture.exception());
        }
        return;
    }
    
    static <V, A> void invokeIndirectly(PendingFuture<V, A> pendingFuture) {
        assert !(!pendingFuture.isDone());
        CompletionHandler<V, Object> handler = pendingFuture.handler();
        if (handler != null) {
            invokeIndirectly(pendingFuture.channel(), handler, pendingFuture.attachment(), pendingFuture.value(), pendingFuture.exception());
        }
        return;
    }
    
    static {
        maxHandlerInvokeCount = (int)AccessController.doPrivileged(new GetIntegerAction("sun.nio.ch.maxCompletionHandlersOnStack", 16));
        myGroupAndInvokeCount = new ThreadLocal<Invoker.GroupAndInvokeCount>() {
            @Override
            protected Invoker.GroupAndInvokeCount initialValue() {
                return null;
            }
        };
    }
    
    static class GroupAndInvokeCount
    {
        private final AsynchronousChannelGroupImpl group;
        private int handlerInvokeCount;
        
        GroupAndInvokeCount(AsynchronousChannelGroupImpl group) {
            this.group = group;
        }
        
        AsynchronousChannelGroupImpl group() {
            return this.group;
        }
        
        int invokeCount() {
            return this.handlerInvokeCount;
        }
        
        void setInvokeCount(int handlerInvokeCount) {
            this.handlerInvokeCount = handlerInvokeCount;
        }
        
        void resetInvokeCount() {
            this.handlerInvokeCount = 0;
        }
        
        void incrementInvokeCount() {
            ++this.handlerInvokeCount;
        }
    }
}
